from .apibase import APIBase, APISimpleDataBase

class BuildLog(APIBase):
    pass

class Build(APIBase):
    
    _aliasmap = {
        'image_tag': 'docker_repo_tag',
        'branch_or_tag_name': 'code_repo_type_value',
        'id': 'build_id'
    }
    
    _hideset = {'docker_repo_tag', 'code_repo_type_value', 'code_repo_type', 'build_id'}
    
    def __init__(self, alauda, data):
        self._alauda = alauda
        super().__init__(data)
        
    @property
    def is_breach(self):
        '指定是分支还是标签。前者返回True，后者返回False'
        return self.data['code_repo_type'] == 'branch'
       
    @property 
    def code_branch(self):
        return self.branch_or_tag_name if self.is_breach else None
            
    @property 
    def code_tag(self):
        return self.branch_or_tag_name if not self.is_breach else None
        
    def __repr__(self):
        return '<Build of {}:{} [{}]>'.format(self.docker_repo_path, self.image_tag, self.id)
        
    def logs(self, start_time = None, end_time = None):
        url = '/v1/builds/{}/logs'.format(self.id)
        r = self._alauda._request_helper(url, 'get')
        if r.status_code == 201:
            return r.json()
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))


from .apibase import APIBase
from .service import Service
from .application import Application


class Node(APIBase):
        
    def __repr__(self):
        if 'total_cpus' in self.resources:
            return '<Node {}: {} ({}core/{})>'.format(self.type, self.private_ip, self.resources['total_cpus'], self.resources['total_mem'])
        else:
            return '<Node {}: {}>'.format(self.type, self.private_ip)
            
class Region(APIBase):
    '''
    可用区
    单个账号中有多个可用区，一些是系统提供的可用区，一些是私有区。
    可用区影响服务和应用的可见性。镜像仓库和构建不受可用区影响。
    '''
    
    
    @classmethod
    def list(cls, alauda):
        '列出应用'
        url = '/v1/regions/{namespace}'
        r = alauda._request_helper(url, 'get')
        
        if 200 == r.status_code:
            ret = []
            for data in r.json():
                ret.append(cls(alauda, data))
            return ret
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
       
    
    _aliasmap = {
    }
    
    _hideset = {}
    
    def __init__(self, alauda, data):
        self._alauda = alauda
        super().__init__(data)
        
    def __repr__(self):
        return '<Region {}:{} [{}]>'.format(self.name, self.display_name, self.id)
        
    def list_node(self):
        url = '/v1/regions/{{namespace}}/{region_name}/nodes'.format(region_name = self.name)
        r = self._alauda._request_helper(url, 'get')
        if 200 == r.status_code:
            ret = []
            for data in r.json():
                ret.append(Node(data))
            return ret
        else:
            raise Exception( '发生了异常：\n{}\n{}'.format(r.status_code, r.text))
        
    def create_service(self, config):
        return Service.create(self._alauda, config)
        
    def get_service(self, name, app_name = None):
        return Service.get(self._alauda, name, app_name, self.name)
    
    def list_service(self, app_name = None):
        return Service.list(self._alauda, app_name, self.name)
        
    def delete_service(self, name, app_name = None):
        return Service.delete_service(self._alauda, name, app_name, self.name)
        
    def create_application(self, name, yml = None):
        return Application.create(self._alauda, name, self.name, yml)
        
    def get_application(self, name):
        return Application.get(self._alauda, name, self.name)
        
    def list_application(self):
        return Application.list(self._alauda, self.name)



import requests
import json
from datetime import datetime

from .region import Region
from .service import Service
from .application import Application
from .repository import Repository


class Alauda:
    
    
    urlbase='https://api.alauda.cn'
    
    debug = False
    
    @staticmethod
    def get_token(username, password, urlbase = None):
        '使用用户名和密码获取Token'
        if urlbase is None:
            urlbase = Alauda.urlbase
            
        r = requests.post(
            url = urlbase + '/v1/generate-api-token',
            headers = {
                'Content-Type': 'application/json',
            },
            data=json.dumps({
                'username': username,
                'password': password
            })
        )
        
        result = json.loads(r.text)
        if 200 == r.status_code:
            return result['token']
        else:
            print('错误，可能是用户名和密码不匹配' + r.text)
            
    def _format_url(self, url):
        return self.urlbase + url.format(namespace = self.namespace)
            
    def _request_helper(self, url, method, data = None, params = None, headers = None, files = None, debug = False):
        '''
        发送请求，将返回值解为json对象
        args:
            - url
            - method http方法
            
        returns:
            - json对象
            - http响应码
        '''
        
        if headers is None:
            headers={}
        headers['Authorization'] = 'Token ' + self.token
        
        if isinstance(data, dict):
            headers['Content-Type'] = 'application/json'
            data = json.dumps(data)
            
        if debug or self.debug:
            r = requests.request(method, 'https://echo.luckymarmot.com/' + self._format_url(url),
                headers = headers, params = params, data = data, files = files);
            filename = datetime.now().strftime('%Y-%m-%d %H_%M_%S.%f') + '.html'
            open(filename, 'w').write(r.text)
            print('请求写入了文件：', filename)
            
        # print('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        # print(method, self._format_url(url))
        # print(data)
        
        r = requests.request(method, self._format_url(url),
            headers = headers, params = params, data = data, files = files);
        
        return r
    
    def __init__(self, namespace, token, default_region = 'BEIJING1', urlbase = None):
        '''
        args:
            - user 用户名，或者组织名
            - token 灵雀云的API token
            - default_region 默认可用区。在没有提供可用区时使用此可用区。
        '''
        self.namespace = namespace
        self.token = token
        self.default_region = default_region
        if urlbase:
            self.urlbase = urlbase
        
        url = '/v1/auth/{namespace}/profile'
        r = self._request_helper(url, 'get')
        
        if 200 == r.status_code:
            result = r.json()
            if result.get('is_available') is True:
                print('连接成功。上次API使用时间：{}'.format(result.get('api_revoked')))
            else:
                raise Exception('账户尚未激活')
        else:
            raise Exception('账户登入失败。请检查namespace和token是否匹配。')
            
        temp = {}
        self._region_map = temp
        self.regions = Region.list(self)
        for region in self.regions:
            temp[region.name] = region
            temp[region.id] = region
            
    def get_region(self, name_or_id):
        '用区域名字或者私有区id获取region'
        return self._region_map.get(name_or_id)
            
     
    def create_service(self, config):
        return Service.create(self, config)
        
    def get_service(self, name, app_name = None):
        return Service.get(self, name, app_name)
    
    def list_service(self, app_name = None):
        return Service.list(self, app_name)
        
    def delete_service(self, name, app_name = None):
        return Service.delete_service(self, name, app_name)
        
    def create_application(self, name, region = None, yml = None):
        return Application.create(self, name, region, yml)
        
    def get_application(self, name):
        return Application.get(self, name)
        
    def list_application(self):
        return Application.list(self)
    
    def create_repo(self, name, description, is_public = False, build_config = None, full_description = ''):
        return Repository.create(self, name, description, is_public, build_config, full_description)
    
    def get_repo(self, name):
        return Repository.get(self, name)
        
    def list_repo(self):
        return Repository.list(self)


from .apibase import APIBase, APISimpleDataBase
from .build import Build

class TagConfig(APIBase):
    
    _aliasmap = {
        'image_tag': 'docker_repo_tag',
        'branch_or_tag_name': 'code_repo_type_value'
    }
    
    _hideset = {'docker_repo_tag', 'code_repo_type_value', 'code_repo_type'}
        
    @property
    def is_breach(self):
        '指定是分支还是标签。前者返回True，后者返回False'
        return self.data['code_repo_type'] == 'branch'
       
    @property 
    def code_branch(self):
        return self.branch_or_tag_name if self.is_breach else None
            
    @property 
    def code_tag(self):
        return self.branch_or_tag_name if not self.is_breach else None
    
    @staticmethod
    def create(image_tag, code_branch = None, code_tag = None, 
    branch_or_tag_name = None, is_breach = None, build_context_path = '/', dockerfile_location = None, 
    is_active = False, build_cache_enabled = False, version_tagging = 0):
        '''
        根据参数创建一个TagConfig
        可以指定 branch_or_tag_name 和 is_breach；
        或者指定 code_branch、code_tag之一。
        '''
        if code_branch:
            if code_tag or branch_or_tag_name:
                raise Exception('code_branch、code_tag、branch_or_tag_name 不能同时使用')
            branch_or_tag_name = code_branch
            is_breach = True
        elif code_tag:
            if branch_or_tag_name:
                raise Exception('code_branch、code_tag、branch_or_tag_name 不能同时使用')
            branch_or_tag_name = code_tag
            is_breach = False
            
        if branch_or_tag_name is None:
            raise Exception('必须指定三者之一：code_branch、code_tag、branch_or_tag_name')
            
        if not dockerfile_location:
            dockerfile_location = build_context_path
        
        if dockerfile_location.find(build_context_path) != 0:
            raise Exception('dockerfile_location 必须以 build_context_path 作为前缀')
            
        
        data = {
            'docker_repo_tag': image_tag,
            'code_repo_type': 'branch' if is_breach else 'tag',
            'code_repo_type_value': branch_or_tag_name,
            'build_context_path': build_context_path,
            'dockerfile_location': dockerfile_location,
            'is_active': is_active,
            'build_cache_enabled': build_cache_enabled,
            'version_tagging': version_tagging,
        }
        
        return TagConfig(data)

class BuildConfig(APIBase):
    
    _aliasmap = {
    }
    
    _hideset = {}
        
    def _update_attributes(self, data):
        self.tag_configs = []
        for c in data['tag_configs']:
            self.tag_configs.append(TagConfig(c))
        
    @property
    def use_cn_node(self):
        return self.build_node == 'cn'
    
    @staticmethod
    def create_client(client, namespace, name, tagconfig,
    use_cn_node = False, email_enabled = False):
        '''
        创建已绑定的代码托管服务的BuildConfig。
        client: 代码托管服务，可选择GitHub, Bitbucket, OSChina, GitCafe
        namespace: 服务的用户名/组织名
        name: 仓库名
        '''
        if isinstance(tagconfig, TagConfig):
            tagconfig = [tagconfig]
        
        data = {
            'code_repo_client': client,
            'code_repo_namespace': namespace,
            'code_repo_name': name,
            'build_node': 'cn' if use_cn_node else 'in',
            'email_enabled': email_enabled,
            'tag_configs': [i.json_data for i in tagconfig]
        }
        
        return BuildConfig(data)
        
    @staticmethod
    def create_simple(url, tagconfig, use_cn_node = False, email_enabled = False):
        '根据版本仓库url创建BuildConfig'
        if isinstance(tagconfig, TagConfig):
            tagconfig = [tagconfig]
        
        data = {
            'code_repo_client': 'Simple',
            'code_repo_clone_url': url,
            'build_node': 'cn' if use_cn_node else 'in',
            'email_enabled': email_enabled,
            'tag_configs': [i.json_data for i in tagconfig]
        }
        
        return BuildConfig(data)

class Repository(APISimpleDataBase):
    
    _aliasmap = {
        'name':'repo_name',
        'build_config_':'build_config'
    }
    
    _hideset = {'service_name'}
    
    @classmethod
    def format_url(cls, alauda, name, url = ''):
        return '/v1/repositories/{namespace}/{name}/{url}'.format(
            name = name, url = url, namespace = alauda.namespace)
            
    @classmethod
    def create(cls, alauda, name, description, 
    is_public = False, build_config = None, full_description = ''):
        data = {
            'repo_name': name,
            'description': description,
            'namespace': alauda.namespace,
            'is_public': is_public
        }
        if build_config:
            data['build_config'] = build_config.json_data
            data['full_description'] = full_description
        print(__import__('json').dumps(data))
        r = alauda._request_helper('/v1/repositories/{namespace}', 'post', data)
        if r.status_code == 201:
            return Repository(alauda, r.json())
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
    
    @classmethod
    def get_data_by_name(cls, alauda, name):
        r = alauda._request_helper(cls.format_url(alauda, name), 'get')
        if r.status_code == 200:
            return r.json()
        elif r.status_code == 404:
            return None
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
            
    @classmethod
    def get(cls, alauda, name):
        data = cls.get_data_by_name(alauda, name)
        if data is None:
            return None
        return cls(alauda, data)
    
    @classmethod
    def list(cls, alauda):
        url = '/v1/repositories/{namespace}'
        r = alauda._request_helper(url, 'get')
        
        if 200 == r.status_code:
            ret = []
            for data in r.json()['results']:
                ret.append(cls(alauda, data, True))
            return ret
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
        
    
    def __init__(self, alauda, data, is_simple = False):
        self._alauda = alauda
        self._build_config = None
        super().__init__(data, is_simple)
        
        
    def _get_full(self):
        return self.get_data_by_name(self._alauda, self.name)
        
    @property
    def api_url(self):
        return self.format_url(self._alauda, self.name)
        
    @property
    def url(self):
        '返回镜像仓库路径'
        url = self._alauda.urlbase.replace('https://api', 'index')
        url = '{}/{}/{}'.format(url, self._alauda.namespace, self.name)
        return url
        
    @property
    def build_config(self):
        if not self.is_automated:
            return None
        if self._build_config is None:
            if self._is_simple:
                self._json_data = self._get_full()
                self._is_simple = True
            self._build_config = BuildConfig(self.build_config_)
        return self._build_config
        
    def __repr__(self):
        return '<Repository [{}]>'.format(self.name)
        
    def delete(self):
        r = self._alauda._request_helper(self.format_url(self._alauda, self.name), 'delete')
        if r.status_code == 204:
            return True
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
        
    def build(self, tag = None, code_commit_id = None):
        '仅构建仓库可以使用此方法。仅当只有一个标签的时候可以省略标签。'
        build_config = self.build_config
        if build_config is None:
            raise Exception('当前仓库不是一个构建仓库')
        if tag is None:
            if len(build_config.tag_configs) == 1:
                tag = build_config.tag_configs[0].image_tag
            else:
                raise Exception('当前仓库具有超过一个构建标签，请显示指定构建标签。')
        
        url = '/v1/builds'
        data = {
            'namespace': self._alauda.namespace,
            'repo_name': self.name,
            'tag': tag
        }
        if code_commit_id is not None:
            data['code_commit_id'] = code_commit_id
        r = self._alauda._request_helper(url, 'post', data)
        if r.status_code == 201:
            return Build(self._alauda, r.json())
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
        



from .apibase import APIBase, APISimpleDataBase
import yaml


class EndPoint(APIBase):
    
    _aliasmap = {
    }
    
    _hideset = {'protocol'}
    
    def __init__(self, data):
        super().__init__(data)
        
    @property
    def domain(self):
        return self._json_data.get('customer_domain') or self.default_domain
        
    @property
    def protocol(self):
        if self.endpoint_type == 'http-endpoint':
            return 'http'
        else:
            return self._json_data['protocol']
        
    @property
    def url(self):
        if self.endpoint_type == 'http-endpoint' and self.service_port == 80:
            return 'http://' + self.domain
            
        return '{}://{}:{}'.format(self.protocol, self.domain, self.service_port)
        
    def __repr__(self):
        return '<EndPoint {}>'.format(self.url)
    

def format_url(alauda, region_name, name, app_name, act = ''):
    url = '/v1/services/{namespace}/{name}/{act}?region_name={region_name}'.format(
        name = name, act = act, namespace = alauda.namespace,
        region_name = region_name if region_name != None else alauda.default_region,
        )
    if app_name is not None:
        url += '&application=' + app_name
    return url

class Service(APISimpleDataBase):
    '''
    灵雀云服务
    '''
    
    _aliasmap = {
        'name':'service_name',
        # 'app_name':'application',
        # 'region_name':['region', 'name']
    }
    
    _hideset = {'service_name', 'application', 'app_name'}
    
    @classmethod
    def create(cls, alauda, config):
        '根据配置文件创建服务'
        
        url = '/v1/services/{namespace}/'
        r = alauda._request_helper(url, 'post', data = config)
        
        if 204 == r.status_code:
            print('成功创建了服务')
            return cls(alauda, r.json())
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
            
    @classmethod
    def get_data(cls, alauda, name, app_name = None, region_name = None):
        url = format_url(alauda, region_name, name, app_name)
        r = alauda._request_helper(url, 'get')
        if 200 == r.status_code:
            return r.json()
        elif r.status_code in (403, 404):
            return None
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
            
    @classmethod
    def get(cls, alauda, name, app_name = None, region_name = None):
        data = cls.get_data(alauda, name, app_name, region_name)
        if data is None:
            return None
        return cls(alauda, data)
    
    @classmethod
    def list(cls, alauda, app_name = None, region_name = None):
        '''
        列出服务
        '''
        # 获取应用中的服务：
        if app_name:
            url = '/v1/applications/{namespace}/{app_name}?region_name={region_name}'.format(
                namespace = alauda.namespace,app_name = app_name, 
                region_name = region_name if region_name != None else alauda.default_region,
                )
            r = alauda._request_helper(url, 'get')
            if 200 == r.status_code:
                ret = []
                for data in r.json()['services']:
                    data['application'] = app_name
                    ret.append(cls(alauda, data, True))
                return ret
            else:
                raise Exception( '发生了异常：\n{}\n{}'.format(r.status_code, r.text))
        
        #获取不在应用中的服务：
        
        url = format_url(alauda, region_name, '|', None)
        url = url.replace('/|/', '/')
        r = alauda._request_helper(url, 'get')
        if 200 == r.status_code:
            ret = []
            for data in r.json()['results']:
                if data.get('application'): #公有区不带application参数会返回包括应用内服务的所有服务。在这里将其过滤。
                    continue 
                ret.append(cls(alauda, data, True))
            return ret
        else:
            raise Exception( '发生了异常：\n{}\n{}'.format(r.status_code, r.text))
    
    @staticmethod
    def start_service(alauda, name, app_name, region_name = None):
        url = format_url(alauda, region_name, name, app_name, 'start')
        r = alauda._request_helper(url, 'put')
        if 204 == r.status_code:
            return True
        elif 400 == r.status_code:
            return False
        else:
            raise Exception( '发生了异常：\n{}\n{}'.format(r.status_code, r.text))
            
    @staticmethod
    def stop_service(alauda, name, app_name, region_name = None):
        url = format_url(alauda, region_name, name, app_name, 'stop')
        r = alauda._request_helper(url, 'put')
        if 204 == r.status_code:
            return True
        elif 400 == r.status_code:
            return False
        else:
            raise Exception( '发生了异常：\n{}\n{}'.format(r.status_code, r.text))
            
    @staticmethod
    def delete_service(alauda, name, app_name, region_name = None):
        url = format_url(alauda, region_name, name, app_name)
        r = alauda._request_helper(url, 'delete')
        if 204 == r.status_code:
            return True
        else:
            raise Exception( '发生了异常：\n{}\n{}'.format(r.status_code, r.text))
            
    
    
    @staticmethod
    def yml_to_json(alauda, yml, region_name = None, application = None, run = False):
        '''
        将yaml格式的服务初始化参数转换为json格式。
        因为yaml可以包含多个服务描述，所以返回结果是一个包含数个json对象的数组。
        args:
            - yml yml对象，或者yml文本，或者打开的文件。
            - region_name 区域名
        returns:
            - json对象序列
        '''
        if not isinstance(yml,dict):
            yml = yaml.load(yml)
        print(yml)
        
        if region_name is None:
            region_name = alauda.default_region
        
        ret = []
        for (service_name,service_data) in yml.items():
            json = {
                'service_name':service_name,
                'namespace': alauda.namespace,
                'application': application,
                'region_name': region_name,
                
                'instance_ports': [],
                'volumes': [],
                'instance_envvars': {},
                
                # 'scaling_mode': 'MANUAL',
                # 'autoscaling_config': {
                #     'metric_name' : 'CPU_UTILIZATION',
                #     'metric_stat' : 'MEAN',
                #     'upper_threshold' : '0.6',
                #     'lower_threshold' : '0.1',
                #     'decrease_delta' : '1',
                #     'increase_delta' : '1',
                #     'minimum_num_instances' : '1',
                #     'maximum_num_instances' : '6',
                #     'wait_period' : '30'
                # },
                }
            ret.append(json)
            
            if run:
                json['target_state'] = 'STARTED'
            else:
                json['target_state'] = 'STOPPED'
                
                
            if 'image' in service_data:
                temp = (service_data['image'] + ':latest').split(':')
                json['image_name'] = temp[0]
                json['image_tag'] = temp[1]
                del service_data['image']
                
            if 'command' in service_data:
                json['run_command'] = service_data['command']
                del service_data['command']
            
            if 'links' in service_data:
                temp = {}
                json['linked_to_apps'] = temp
                for l in service_data['links']:
                    l = l.split(':')
                    if len(l) == 2:
                        temp[l[0]] = l[1]
                    else:
                        temp[l[0]] = l[0]
                del service_data['links']
                
            if 'size' in service_data:
                json['instance_size'] = service_data['size']
                del service_data['size']
                
            if 'number' in service_data:
                json['target_num_instances'] = service_data['number']
                del service_data['number']
            
            if 'ports' in service_data:
                for p in service_data['ports']:
                    p = p.split('/')
                    container_port = p[0]
                    protocol = 'tcp'
                    if 2 == len(p):
                        protocol = p[1]
                    else:
                        if 80 == container_port or 8080 == container_port:
                            protocol == 'http'
                    temp = {
                        'container_port':container_port,
                        'protocol': protocol
                        }
                    if 'http' == protocol:
                        temp['endpoint_type'] = 'http-endpoint'
                    elif 'tcp' == protocol:
                        temp['endpoint_type'] = 'tcp-endpoint'
                    else:
                        raise Exception('指定了不支持的端口协议：{}'.format(protocol))
                    json['instance_ports'].append(temp)
                    
                del service_data['ports']
            
            if 'expose' in service_data:
                for p in service_data['expose']:
                    json['instance_ports'].append({
                            'container_port': p,
                            'protocol': 'tcp',
                            'endpoint_type': 'internal-endpoint'
                        })
                del service_data['expose']
            
            if 'volumes' in service_data:
                for v in service_data['volumes']:
                    v = (v + ':10').split(':')
                    json['volumes'].append({
                        'app_volume_dir' : v[0],
                        'volume_type' : 'EBS',
                        'size_gb' : v[1],
                        })
                del service_data['volumes']
                
            if 'environment' in service_data:
                envs = service_data['environment']
                if isinstance(envs, list):
                    for env in envs:
                        env = env.split('=')
                        if 1 == len(env):
                            json['instance_envvars'][env[0]] = None
                        else:
                            json['instance_envvars'][env[0]] = env[1]
                else:
                    for (k, v) in envs.items():
                        json['instance_envvars'][k] = v
                        
                del service_data['environment']
                
            if len(service_data.items()) > 0:
                print('无法处理的参数:\n{}'.format(service_data))
        
        return ret
    
    def __init__(self, alauda, data, is_simple = False):
        self._alauda = alauda
        self._endpoints = None
        super().__init__(data, is_simple)
        
    def list_endpoint(self):
        if not self._endpoints:
            self._endpoints = []
            if 'alauda_load_balancer' in self._json_data:
                for data in self._json_data['alauda_load_balancer'].get('endpoints', []):
                    self._endpoints.append(EndPoint(data))
            else:
                if 'instance_ports' not in self._json_data:
                    self._update_to_full()
                for data in self._json_data.get('instance_ports', []):
                    self._endpoints.append(EndPoint(data))
        return self._endpoints
        
    @property
    def app_name(self):
        return self._json_data.get('application')
        
    @property
    def region_name(self):
        region_id = self._json_data.get('region_id')
        if region_id:
            return self._alauda.get_region(region_id).name
        else:
            return self._json_data['region']['name']
        
    def _get_full(self):
        return self.get_data(self._alauda, self.name, self.app_name, self.region_name)
    
    def _format_url(self, url = ''):
        return format_url(self._alauda, self.region_name, self.name, self.app_name, url)
        
    @property
    def api_url(self):
        return self._format_url()
            
    def start(self):
        '启动此服务'
        return self.start_service(self._alauda, self.name, self.app_name)
    
    def stop(self):
        '停止此服务'
        return self.stop_service(self._alauda, self.name, self.app_name)
            
    def delete(self):
        '删除此服务'
        return self.delete_service(self._alauda, self.name, self.app_name)
            
    def update(self, name = None, target_num_instances = None, \
    image_tag = None, scaling_mode = None):
        '更新服务参数'
        data = {
            'service_name':name or self.name,
            'target_num_instances': target_num_instances or self.target_num_instances,
            'image_tag': image_tag or self.image_tag,
            'scaling_mode': scaling_mode or self.scaling_mode
        }
        r = self._alauda._request_helper(self.api_url, 'put', data)
        if 204 == r.status_code:
            return True
        else:
            raise Exception( '发生了异常：\n{}\n{}'.format(r.status_code, r.text))
    
    def __repr__(self):
        if self.app_name is None:
            return '<Service [{}] in [{}]>'.format(self.name, self.region_name)
        else:
            return '<Service [{}] of [{}] in [{}]>'.format(self.name, self.app_name, self.region_name)


import json
'''
参考自 https://github.com/sigmavirus24/github3.py/blob/develop/github3/models.py
'''
class APIBase(object):
    '用于映射JSON API对象的基类'
    
    _aliasmap = {}
    '''
    别名映射表
    如果令 _aliasmap['attr'] = 'other'，则：
    访问 apibase.attr 返回 apibase._json_data['other']
    访问 apibase.other 也会返回 apibase._json_data['other']
    如果令 _aliasmap['attr'] = ['lv1','lv2']，则：
    访问 apibase.attr 返回 apibase._json_data['lv1']['lv2']
    '''
    
    _hideset = set()
    '''
    属性隐藏表
    如果 'attr' in _hideset，则：
    访问 apibase.attr 不会返回 apibase._json_data['attr']
    '''
    
    def __init__(self, data):
        if isinstance(data, str):
            data = json.loads(data)
        self._update_attributes(data)
        self._json_data = data
        self.__dir_cache__ = None
        
    def _update_attributes(self, data):
        '子类的初始化更新方法'
        pass
        
    @property
    def json_data(self):
        '返回原始json对象'
        return self._json_data
        
    def __getattr__(self, attribute):
        '令json对象的属性可以直接访问'
        if attribute in self._aliasmap:
            attribute = self._aliasmap[attribute]
            if not isinstance(attribute, list):
                attribute = [attribute]
            temp = self._json_data
            for key in attribute:
                if key in temp:
                    temp = temp[key]
                else:
                    raise AttributeError(attribute)
            return temp
                
        if attribute not in self._hideset and attribute in self._json_data:
            return self._json_data[attribute]
        else:
            print('发生错误', attribute, self._json_data)
            raise AttributeError(attribute)
    
    def __dir__(self):
        '令dir命令枚举属性时包含json对象的属性'
        def _in(attribute):
            if isinstance(attribute, list):
                temp = self._json_data
                for key in attribute:
                    if key in temp:
                        temp = temp[key]
                    else:
                        return False
                return True
            else:
                return attribute in self._json_data
                
        if self._json_data is None:
            return super().__dir__()
        if self.__dir_cache__ is None:
            # 排除隐藏的属性
            datakeys = set(self._json_data.keys())
            temp =  datakeys - self._hideset
            # 添加属性别名
            temp |= {k for k, v in self._aliasmap.items() if _in(v)} 
            temp |= set(super().__dir__())
            self.__dir_cache__ = list(temp)
        return self.__dir_cache__


class APILiteBase(APIBase):
    '''
    在使用RESTful API时，可以获取单个对象或对象列表。
    有时，对象列表中的对象相比单个对象来说仅具有一部分关键信息。
    APILiteBase类可以使用该关键信息创建代理。
    访问代理具有的关键信息时，代理直接返回；
    当访问代理中没有的属性或方法时，代理先请求数据对象的完整版本，然后将请求转发给该完整版本
    '''
    _full_data = None
    
    def _get_full(self):
        '实现此方法以获取数据对象的完整版本'
        pass
        
    def __dir__(self):
        if self._full_data is None:
            self._full_data = self._get_full()
        return dir(self._full_data)
    
    def __getattr__(self, attribute):
        if self._full_data is not None:
            return getattr(self._full_data, attribute)
        try:
            return super().__getattr__(attribute)
        except AttributeError:
            self._full_data = self._get_full()
            return getattr(self._full_data, attribute)
            
class APISimpleDataBase(APIBase):
    '''
    允许传入简化版json数据。
    当持有简化版数据，且访问了该数据中不存在的字段时，
    首先请求完整数据，然后返回完整数据中的对应字段。
    '''
    
    def _get_full(self):
        '实现此方法以获取数据对象的完整版本'
        pass
    
    def _update_to_full(self):
        if not self._is_simple:
            return
        self._json_data = self._get_full()
        self._is_simple = False
    
    def __init__(self, data, is_simple = False):
        self._is_simple = is_simple
        super().__init__(data)
        
    def __dir__(self):
        if self._is_simple:
            self._update_to_full()
        return super().__dir__()
    
    def __getattr__(self, attribute):
        if self._is_simple:
            try:
                return super().__getattr__(attribute)
            except AttributeError:
                self._update_to_full()
            
        return super().__getattr__(attribute)
                
            


from .apibase import APIBase
class Application(APIBase):
    '''
    灵雀云应用
    '''
    
    _aliasmap = {
        'region_id': 'region',
        'name':'app_name',
    }
    
    _hideset = {'app_name', 'region'}
    
    @classmethod
    def create(cls, alauda, name, region_name = None, yml = None):
        url = '/v1/applications/{namespace}'
        data = {
           'app_name': name,
           'region': region_name if region_name else alauda.default_region,
           'namespace': alauda.namespace,
        }
        if yml:
            data['services'] = yml
            
        r = alauda._request_helper(url, 'post', data = data)
        
        if 201 == r.status_code:
            return cls(alauda, r.json())
        else:
            print(data)
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
            
    @classmethod
    def get(cls, alauda, name, region_name = None):
            
        url = '/v1/applications/{namespace}/{name}'.format(
            name = name, namespace = alauda.namespace)
        #region_name可以为空字符串。这代表使用系统提供的所有region。
        params = {'region': region_name if region_name != None else alauda.default_region}
        r = alauda._request_helper(url, 'get', params = params)
        
        if 200 == r.status_code:
            return cls(alauda, r.json())
        elif 404 == r.status_code:
            return None
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
    
    @classmethod
    def list(cls, alauda, region_name = None):
        '列出应用'
        url = '/v1/applications/{namespace}'
        #region_name可以为空字符串。这代表使用系统提供的所有region。
        params = {'region': region_name if region_name != None else alauda.default_region}
        r = alauda._request_helper(url, 'get', params = params)
        
        if 200 == r.status_code:
            ret = []
            for data in r.json():
                ret.append(cls(alauda, data))
            return ret
        else:
            print(url, params)
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
    
    def __init__(self, alauda, data):
        self._alauda = alauda
        super().__init__(data)
        
    @property
    def region(self):
        return self._alauda.get_region(self.region_id)
        
    @property
    def region_name(self):
        return self.region.name
    
    def _format_url(self, url = ''):
        return '/v1/applications/{namespace}/{name}/{url}?region={region}'.format(
            name = self.name, url = url, namespace = self._alauda.namespace, region = self.region_name)

    @property
    def api_url(self):
        return self._format_url()
        
    @property
    def yaml(self):
        url = self._format_url('/yaml')
        r = self._alauda._request_helper(url, 'get')
        if 200 == r.status_code:
            return r.text
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
            
    def update(self, yaml):
        files = {'services': yaml}
        r = self._alauda._request_helper(self.api_url, 'put', files = files)
        if 204 == r.status_code:
            return
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
            
    def list_service(self):
        return self.region.list_service(self.name)
        
    def delete_service(self, name):
        return self.region.delete_service(name, self.name)
        
    def get_service(self, name):
        return self.region.get_service(name, self.name)
            
            
    def start(self):
        url = self._format_url('/start')
        r = self._alauda._request_helper(url, 'put')
        if 204 == r.status_code:
            return r.text
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
            
    def stop(self):
        url = self._format_url('/start')
        r = self._alauda._request_helper(url, 'put')
        if 204 == r.status_code:
            return r.text
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
            
    def delete(self):
        r = self._alauda._request_helper(self.api_url, 'delete')
        if 204 == r.status_code:
            return r.text
        else:
            raise Exception('发生了异常：\n{}\n{}'.format(r.status_code, r.text))
            
    
    
    def __repr__(self):
        return '<Application [{}] in [{}], len[{}]>'.format(self.name, self.region_name, len(self.services))
